// implements the newcamd protocol (versions 5.20 & 5.25)

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/poll.h>

#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

#include "newcamd.h"
#include "des.h"
#include "newcamd-srv.h"


// ------------------------------- newcamd client class member implementation ------------------------------


newcamd_srv::newcamd_srv(char *server_url, unsigned int timeout, protocol_version_t version, int fdsock, unsigned char *xcdr, int xcdrsize)
{
	char config_key[29];
	unsigned char config_key_byte_count = sizeof(des_key);
	int des_key_byte;

	set_protocol_version( version );
	
	status = 0;
	ready = false;
	network_timeout = timeout;
	fd=fdsock;
	
	sent_sequence_id = 0;
	received_sequence_id = 0;

	memset(cw_0, 0, sizeof(cw_0));
	memset(cw_1, 0, sizeof(cw_1));
	
	strcpy(url, server_url);
	cdrsize=xcdrsize;
	memcpy(cdrbuffer,xcdr,xcdrsize);

	memset(username, 0, sizeof(username));
	memset(password, 0, sizeof(password));
	memset(random_key, 0, sizeof(random_key));
	memset(des_key, 0, sizeof(des_key));
	memset(login_key, 0, sizeof(login_key));
	memset(session_key, 0, sizeof(session_key));
	
	memset(expanded_login_key, 0, sizeof(expanded_login_key));
	memset(expanded_session_key, 0, sizeof(expanded_session_key));


	//NCD: user pass 0011223344556677665544332211 
	printf ("%s\n",url);
	sscanf(url, "NCD: %4x %s %s %s %s",&camtype,config_key, username, password, config_key);
	printf ("User:%s  Pass:%s  Key:%s\n",username,password,config_key);


	while( config_key_byte_count-- )
	{
		sscanf(&config_key[config_key_byte_count * 2], "%02X", &des_key_byte);
		des_key[config_key_byte_count] = (unsigned char)des_key_byte;
	}

	debug_output(cdrbuffer,cdrsize,"NCD Card data request :");
	card_caid= cdrbuffer[1]*256 + cdrbuffer[2];
	//printf("CARD CAID: %04x\n",card_caid);
	server_login();
	if (status) 
		{
		printf("Not Successful Login..\n");
		return;
		}
	ready = true;
	dbg = 0;
}

newcamd_srv::~newcamd_srv( void )
{
	 if (fd>=0)
		{
		if (shutdown(fd, 2)<0) perror("socket shutdown");
		else if (close(fd)<0) perror("socket close");
		}
}

bool newcamd_srv::is_ready( void )
{
	return ready;
}

void newcamd_srv::set_protocol_version( protocol_version_t version )
{
	protocol_version = version;
	switch( version )
	{
		case VERSION_520:
			custom_data_length = 4;
			break;
			
		case VERSION_525:
			custom_data_length = 8;
			break;
		
		default:
			custom_data_length = 8;
			break;
	}
	
	set_sid(camtype);    // x0666   //0x8888 oscams   //4343 cccam
}

void newcamd_srv::set_sid( unsigned short int sid )
{
	memset(custom_data, 0, sizeof(custom_data));
	switch( protocol_version )
	{
		case VERSION_520:
			custom_data[2] = sid >> 8;
			custom_data[3] = sid & 0xFF;
			break;
			
		case VERSION_525:
			custom_data[0] = sid >> 8;
			custom_data[1] = sid & 0xFF;
			break;
		
		default:
			custom_data[0] = sid >> 8;
			custom_data[1] = sid & 0xFF;
			break;
	}
}

	

void newcamd_srv::server_login( void )
{
	unsigned char buffer[CWS_NETMSGSIZE];
	unsigned char bufferr[CWS_NETMSGSIZE];
	memset(buffer, 0, CWS_NETMSGSIZE);
	
	// ----------------------------------------- LOGIN --------------------------------------------------
	//build login command data
	char *encrypted_password = crypt(password,"$1$abcdefgh$");
	unsigned short int username_length = strlen(username);
	unsigned short int password_length = strlen(encrypted_password);
  	memcpy(buffer, username, username_length + 1);
  	memcpy(buffer + username_length + 1, encrypted_password, password_length + 1);
	

	//server should initiate sequence with 14-byte random key. get it and create the login key
	if( network_write(random_key, sizeof(random_key)) != sizeof(random_key) )
	{
		status = 1;
		printf( "[Newcamd_SRV] Failed to receive random server key.");
		return;
	}
		
	create_login_key(); //xor random key with common des key
	memcpy(des.des_key, expanded_login_key, sizeof(des.des_key));
	des.schedule_key();
	
	receive_message(bufferr, NULL, LOGIN_KEY );
	if (status) return;
	if (memcmp(bufferr+3,buffer,username_length+password_length+2 )!=0)
		{
  		status = 1;
		send_command(MSG_CLIENT_2_SERVER_LOGIN_NAK, buffer, 0, NULL, LOGIN_KEY);
		printf("[Newcamd_SRV] Login request was rejected. user/pass incorrect : %s\n",bufferr+3);
  		return;
  		}
	
	send_command(MSG_CLIENT_2_SERVER_LOGIN_ACK, buffer, 0, NULL, LOGIN_KEY);
	if( status ) return;
	printf("[Newcamd_SRV] Correct User and Password\n");

	//login was successful, now create session key
	create_session_key( encrypted_password ); //xor password with common des key
	memcpy(des.des_key, expanded_session_key, sizeof(des.des_key));
	des.schedule_key();
  	
	//request card data
	receive_message(bufferr, NULL, SESSION_KEY );
	if( status ) return;
	if( bufferr[0] != MSG_CARD_DATA_REQ )
  		{
  		status = 1;
  		printf( "[Newcamd_SRV] Card data request failed.\n");
  		return;
  		}
	//printf("Will respond to Card Data Req\n");
	//status=0;
	send_command(MSG_CARD_DATA, cdrbuffer,cdrsize, NULL, SESSION_KEY);
	
	return;
}

void newcamd_srv::Sendecho(unsigned char *keys, int len)
{
	if (dbg) debug_output(keys, len,"NCD TX:");
	send_message( keys,len, NULL, SESSION_KEY);

}

bool newcamd_srv::checkdata( void )
{
	unsigned char 	buffer[512];
	int 		len;
	unsigned int	i;

	len=receive_message(buffer, custom_data, SESSION_KEY );
	//printf("Msg len : %d\n",len);
	if ((len<3) || (len>CWS_NETMSGSIZE))
		{ //if (dbg) printf("Server : invalid message size %d\n",len);
		//if (status > 1) ready=false;
		return false; }

	if (!( (buffer[0]==0x80)  || (buffer[0]==0x81) || (buffer[0]==MSG_KEEPALIVE) )) // || (buffer[0]==0x82)
		{
		if (dbg) printf("NCD RX: EMM received.\n");
		buffer[1]=0;
		buffer[2]=0;
		Sendecho(buffer,3);
		return false;   //  
		}
	tableid=buffer[0];
	memcpy(ECMbuffer,buffer,len);
	ECMsize=len;
	
	if (buffer[0]==MSG_KEEPALIVE)
		{
		if (dbg) printf("NCD RX: Keepalive received.\n");
		Sendecho(buffer,3);
		return false;
		}


	//debug_output(buffer,len,"ECM:");

	switch (card_caid&0xff00)
		{	
			case 0x500:	if (ECMbuffer[4]==0x90) PROVID=(ECMbuffer[6]<<16)+(ECMbuffer[7]<<8)+(ECMbuffer[8] & 0xF0);
							else 	PROVID=(ECMbuffer[8]<<16)+(ECMbuffer[9]<<8)+(ECMbuffer[10]);
					break;			//VIA
			case 0x600: 	PROVID=0;  break;	//IRDETO HEP OOOO
			case 0x100: 	PROVID=(ECMbuffer[3]<<8)+ECMbuffer[4]; break;			//SECA OK
			case 0x1700: 	PROVID=0; break;						//1702=0
			case 0x1800: 	PROVID=0; break;

			case 0x0900:	PROVID=0; break;
			case 0x0b00:	PROVID=0; break;
			case 0x0d00: 	{ for(int i=8; i<ECMsize; i+=ECMbuffer[i+1]+2) if(ECMbuffer[i]==0x83) PROVID =ECMbuffer[i+2]&0xFC; break; }  //CRYPTOWORKS
			default:	PROVID=0; break;
		}

	SID=(custom_data[0]<<8)+custom_data[1];
	//sprintf(sid,"%04X",SID);
	//if (dbg) printf("\nSID:%04x (%04d)\n",SID,SID);
	

	return true;
}

void newcamd_srv::send_command(net_msg_type_t command, unsigned char *data, unsigned short int data_length, unsigned char *custom_data, key_encryption_t key_type)
{
	unsigned char command_buffer[3 + data_length];
	command_buffer[0] = command;
	command_buffer[1] = (data_length >> 8) & 0x0F;
	command_buffer[2] = data_length & 0xFF;
	if( data ) memcpy(command_buffer + 3, data, data_length);
	
	return send_message(command_buffer, sizeof(command_buffer), custom_data, key_type);
}

net_msg_type_t newcamd_srv::receive_command( unsigned char *custom_data, key_encryption_t key_type )
{
	unsigned char command_buffer[CWS_NETMSGSIZE];
	
	if( receive_message(command_buffer, custom_data, key_type) == 3 )
		return (net_msg_type_t) command_buffer[0];
	else
		return (net_msg_type_t)(-1);
}

int newcamd_srv::receive_message(unsigned char *message, unsigned char *custom_data, key_encryption_t key_type)
{
	unsigned char message_buffer[CWS_NETMSGSIZE];
	unsigned short int message_length;
	int bytes_received,ix;
	
	//message header length is 2-byte size header (unencrypted) + 2-byte sequence id (encrypted) + 
	//8 or 4-byte custom data field (encrypted).
	unsigned short int message_header_length = 2 + 2 + custom_data_length;
	
	//read first 2 bytes, this is the message length
	if( (bytes_received = network_read(message_buffer, 2)) != 2 )
	{
		switch( bytes_received )
		{
			case -1:
				status = 3;
				ready=false;
				printf( "[Newcamd_SRV] Failed to read network stream. Server disconnected.\n");
				return 0;
			
			case -2:
				status = 1;		//timeout 1 olabilir.
				//ready=false;
				//printf( "[Newcamd_SRV] Failed to read network stream. Time-out.\n");
				return 0;
				
			default:
				status = 5;
				ready=false;
				printf( "[Newcamd_SRV] Failed to read network stream.\n");
				return 0;
		}
	}
	
	//message length is big-endian
	message_length = message_buffer[0] << 8 | message_buffer[1];
	
	//check if message length is valid
	if( message_length > CWS_NETMSGSIZE - 2 )
	{
		status = 1;
		printf( "[Newcamd_SRV] Received message too large (%d bytes).\n", message_length + 2);
		return 0;
	}
	
	//read the rest of the message
	if( (bytes_received = network_read(message_buffer + 2, message_length)) != message_length )
	{
		switch( bytes_received )
		{
			case -1:
				status = 3;
				ready=false;
				printf( "[Newcamd_SRV] Failed to read network stream. Server disconnected.\n");
				return 0;
			
			case -2:
				status = 1;   //timeout 1 olabilir.
				//ready=false;
				printf( "[Newcamd_SRV] Failed to read network stream. Time-out.\n");
				return 0;
				
			default:
				status = 5;
				ready=false;
				printf( "[Newcamd_SRV] Failed to read network stream.\n");
				return 0;
		}
	}
	message_length += 2;
	
	
	
	//decrypt the whole message, includes the 2-byte size header, using apropriate key_type	switch( key_type )
	switch( key_type )
	{
		case LOGIN_KEY:
			memcpy(des.des_key, expanded_login_key, sizeof(des.des_key));
			break;
			
		case SESSION_KEY:
			memcpy(des.des_key, expanded_session_key, sizeof(des.des_key));
			break;
	}
	des.decrypt(message_buffer, message_length);
	message_length -= sizeof(i_DES_cblock);
	
	//check sum the result
	if(des.xor_checksum(message_buffer + 2, message_length - 2))
	{
		status = 1;
		printf( "[Newcamd_SRV] Checksum error on decrypted packet.\n");
		return 0;
	}
	
	//get size from 12-bit command data length of message, add 3 bytes (from command header) to get 
	//total size of data to be returned
	unsigned short command_length = 
		((message_buffer[message_header_length + 1] & 0x0F) | message_buffer[message_header_length + 2]) + 3;
	
	//get the received sequence id (used for message synchronization)
	received_sequence_id = (message_buffer[2] << 8) | message_buffer[3];
	
	//if custom_data pointer is present, copy custom data from mesage to it
	if( custom_data ) memcpy(custom_data, message_buffer + 4, custom_data_length);
	
	memcpy(message, message_buffer + message_header_length, command_length);
	ecmcrc=0;
	for (ix=0;ix<command_length;ix++) ecmcrc += message[ix];
	//debug_output(message_buffer, message_length, "< ");
	return command_length;
}

void newcamd_srv::send_message(unsigned char *message, unsigned short int length, unsigned char *custom_data, key_encryption_t key_type)
{
	unsigned char message_buffer[CWS_NETMSGSIZE];
	memset(message_buffer, 0, CWS_NETMSGSIZE);
	
	//message header length is 2-byte size header (unencrypted) + 2-byte sequence id (encrypted) + 
	//8 or 4-byte custom data field (encrypted).
	unsigned short int message_header_length = 2 + 2 + custom_data_length;
	
	//total message length is the message header length + the length of the command data received.
	unsigned short int message_length = message_header_length + length;
	
	if( length < 3 || (message_header_length + length ) > CWS_NETMSGSIZE )
	{
		status = 1;
		ready=false;
		printf( "[Newcamd_SRV] Invalid message size.\n");
		return;
	}
	
	//increment and big-endian set new session id to bytes 2 & 3 of message buffer
	message_buffer[2] = (received_sequence_id >> 8);
	message_buffer[3] = (received_sequence_id) & 0xFF;
	
	memcpy(message_buffer + message_header_length, message, length);
	
	//if custom data is received, embed it in the custom data section of the message
	//depending on protocol version, custom data length can be either 4 bytes or 8 bytes.
	if( custom_data ) memcpy(message_buffer + 4, custom_data, custom_data_length);
	
	//debug_output(message_buffer, message_length, " before encrypt> ");
	
	//pad message for encrypting
	message_length = des.pad_message(message_buffer, message_length);
	
	//3des encrypt full message with appropriate key
	switch( key_type )
	{
		case LOGIN_KEY:
			memcpy(des.des_key, expanded_login_key, sizeof(des.des_key));
			break;
			
		case SESSION_KEY:
			memcpy(des.des_key, expanded_session_key, sizeof(des.des_key));
			break;
	}
	des.schedule_key();
	des.encrypt(message_buffer, message_length, message_buffer);
	
	//compensate for message padding (add 1 DES block to message length)
	message_length += sizeof(i_DES_cblock);
	
  	//set total message length UNENCRYPTED in bytes 0 & 1 of message
  	message_buffer[0] = (message_length - 2) >> 8;
  	message_buffer[1] = (message_length - 2) & 0xFF;
  	
  	//write message to network socket
  	if( network_write(message_buffer, message_length) != message_length )
  	{
  		status = 3;
		ready=false;
  		printf( "[Newcamd_SRV] Error writing network stream.\n");
  		return;
  	}
	//sent_sequence_id++;
	
  	status = 0;
  	//if (dbg) printf( "[Newcamd_SRV] Message sent successfully.\n");
}

void newcamd_srv::create_login_key()
{
	//xors the received 14-byte random key with the 14-byte common key.
	//expands the resulting 14-byte login_key to a 16-byte expanded_login_key
	des.xor_buffers(login_key, random_key, des_key, sizeof(login_key));
	des.expand(login_key, expanded_login_key);
	//debug_output(login_key,14,"\nLogin key :");
	//debug_output(expanded_login_key,16,"Expanded Login key16 :");
}

void newcamd_srv::create_session_key( char *encrypted_password )
{
	//xors the common des key with the authenticaded user password to form the 14-byte session key
	//expands 14-byte sesion_key to a 16-byte expanded_session_key
	memcpy(session_key, des_key, sizeof(session_key));
	for(unsigned int i=0;i<strlen(encrypted_password);i++)
		session_key[i % sizeof(session_key)] ^= encrypted_password[i];
	des.expand(session_key, expanded_session_key);
	//debug_output(expanded_session_key,16,"Exp key16 :");
}


int newcamd_srv::network_write(unsigned char *buffer, unsigned int length)
{
	//write with time-out
	struct pollfd pfd;
	int write_result;
	
	pfd.fd = fd;
	pfd.events = POLLOUT | POLLERR;
	pfd.revents = 0;
	
	if(poll(&pfd, 1, network_timeout) != 1) return -2;
	
	if(pfd.revents & POLLOUT)
	{
		if( (write_result = write(fd, buffer, length)) <= 0 ) ready = false;
		return write_result;
	}
	else
		return -1;
}

int newcamd_srv::network_read(unsigned char *buffer, unsigned int length, int override_timeout)
{
	//read with time-out
	struct pollfd pfd;
	int read_result;
	
	pfd.fd = fd;
	pfd.events = POLLIN | POLLERR;
	pfd.revents = 0;
	
	if( override_timeout < 0 ) override_timeout = network_timeout;
	
	if(poll(&pfd, 1, override_timeout) != 1) return -2;
	
	if(pfd.revents & POLLIN)
	{
		if( (read_result = read(fd, buffer, length)) <= 0 ) ready = false;
		//debug_output(buffer, read_result, "e< ");
		return read_result;
	}
	else
		return -1;
}




void newcamd_srv::debug_output(unsigned char *buffer, unsigned int length, char *header)
{
	printf("%s ", header);
	for(unsigned int i=0;i<length;i++) printf("%02X ", buffer[i]);
	printf("\n");
}


